﻿/*
 * Copyright(c) 2009 cuegraphix.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

 package com.cuegraphix.layouts
{
	import com.cuegraphix.events.LiquidLayoutEvent;
	import flash.display.Sprite;
	import flash.display.DisplayObject;
	import flash.display.Stage;
	import flash.display.StageAlign;
	import flash.display.StageScaleMode;
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.utils.Dictionary;
	import jp.nium.display.ExDocument;
	import jp.nium.events.DocumentEvent;
	import jp.nium.events.EventIntegrator;
	import jp.nium.utils.StageUtil;
	
	/**
	* フルFlashサイト等で、ブラウザ（Stage）のリサイズ時にオブジェクトを定位置もしくは可変サイズにするためのクラスです。
	*
	* Layoutオブジェクトを作成し、addLayoutするだけで、リサイズ時に設定した表示方法＆タイミングで表示を変更できます。
	*
	* @author cue
	*/
	public class LiquidLayout extends EventIntegrator
	{
		/**
		 * @private
		 */
		protected var _document:ExDocument;
		
		/**
		 * @private
		 */
		protected var _listeners:Dictionary;
		
		/**
		 * @private
		 * LiquidLayout.addLayout の config 引数にて使用できるプロパティキー名および
		 * LiquidLayout プロパティ名⇒ DisplayObject プロパティ名対応表
		 */
		protected static var _correspondingProperties:Object = {
			left : "x", center: "x", right : "x",
			top  : "y", middle: "y", bottom: "y",
			centerX: "x", centerY: "y",
			width: "width", height: "height",
			stageWidth: "width", stageHeight: "height",
			scaleX: "scaleX", scaleY: "scaleY"
		};
		
		/**
		 * @private
		 * LiquidLayout 内で使用する予約語（プロパティ名から除外）
		 */
		protected static var _restrictedWords:Object = {
			eventType: true, transitionType: true,
			onResize: true, onResizeStart: true, onResizeProgress: true, onResizeComplete: true,
			onResizeParams: true, onResizeStartParams: true, onResizeProgressParams: true, onResizeCompleteParams: true
		};
		
		/**
		 * @private
		 * 追加拡張プロパティ格納
		 */
		protected static var _specialProperties:Object = {};
		
		/**
		 * @private
		 * イベントタイプ名のリスト
		 */
		protected static var _restrictedEvents:Object = {
			resizeAlways: true, resizeStart: true, resizeProgress: true, resizeComplete: true
		};
		
		/**
		 * @private
		 * パーセント算出プロパティ対応表
		 */
		internal var calcParcentProperties:Object = {
			left : "width", center: "width", right : "width",
			top  : "height", middle: "height", bottom: "height"
		}
		
		/**
		 * addLayout にて追加時 LiquidLayout クラスで定義された各対応プロパティ名以外のプロパティに値を代入したい場合、その LiquidLayout プロパティ名に付与して使用する接尾語を定義します。
		 */
		public static const SUB_PROPERTY_NAME_SUFFIX:String = "PropName";
		
		/**
		 *
		 */
		private var _initialized:Boolean = false;
		
		
		/**
		 * レイアウト上の最小幅値を示す数値です。
		 * @default		0
		 */
		public function get stageMinWidth():int { return _stageMinWidth || ExDocument.stageMinWidth; }
		public function set stageMinWidth( value:int ):void
		{
			_stageMinWidth = value;
		}
		private var _stageMinWidth:int = 0;
		
		/**
		 * レイアウト上の最小高さ値を示す数値です。
		 * @default		0
		 */
		public function get stageMinHeight():int { return _stageMinHeight || ExDocument.stageMinHeight; }
		public function set stageMinHeight( value:int ):void
		{
			_stageMinHeight = value;
		}
		private var _stageMinHeight:int = 0;
		
		/**
		 * レイアウト上の最大幅値を示す数値です。
		 * @default		2880
		 */
		public function get stageMaxWidth():int { return _stageMaxWidth; }
		public function set stageMaxWidth( value:int ):void
		{
			_stageMaxWidth = value;
		}
		private var _stageMaxWidth :int = 2880;
		
		/**
		 * レイアウト上の最大高さ値を示す数値です。
		 * @default		2880
		 */
		public function get stageMaxHeight():int { return _stageMaxHeight; }
		public function set stageMaxHeight( value:int ):void
		{
			_stageMaxHeight = value;
		}
		private var _stageMaxHeight:int = 2880;
		
		/**
		 * 設定されているStageオブジェクトです。
		 * 読み取り専用
		 */
		public function get stage():Stage { return ExDocument.stage; }
		
		/**
		 * ドキュメントファイルに設定されている規定幅です。
		 * 読み込みが完全に完了していない場合は0を返します。
		 * 読み取り専用
		 */
		public function get contentWidth():int { return ExDocument.contentWidth; }
		
		/**
		 * ドキュメントファイルに設定されている規定高さです。
		 * 読み込みが完全に完了していない場合は0を返します。
		 * 読み取り専用
		 */
		public function get contentHeight():int { return ExDocument.contentHeight; }
		
		/**
		 * 現在のステージの幅を返します。
		 * 読み取り専用
		 */
		public function get stageWidth():int { return Math.max( Math.min( ExDocument.stageWidth, stageMaxWidth ), stageMinWidth ); }
		
		/**
		 * 現在のステージの高さを返します。
		 * 読み取り専用
		 */
		public function get stageHeight():int { return Math.max( Math.min( ExDocument.stageHeight, stageMaxHeight ), stageMinHeight ); }
		
		/**
		 * オリジナルのステージの幅を返します。
		 * 読み取り専用
		 */
		public function get originStageWidth():int { return stage ? stage.stageWidth : 0; }
		
		/**
		 * オリジナルのステージの高さを返します。
		 * 読み取り専用
		 */
		public function get originStageHeight():int { return stage ? stage.stageHeight : 0; }
		
		/**
		 * ステージ幅の左端の数値を返します。
		 * 読み取り専用
		 */
		public function get left():Number { return ExDocument.left; }
		
		/**
		 * ステージ幅の中心の数値を返します。
		 * 読み取り専用
		 */
		public function get center():Number { return centerX; }
		
		/**
		 * ステージ幅の中心の数値を返します。
		 * 読み取り専用
		 */
		public function get centerX():Number { return right / 2 + left / 2; }
		
		/**
		 * ステージ幅の右端の数値を返します。
		 * 読み取り専用
		 */
		public function get right():Number { return left + stageWidth; }
		
		/**
		 * ステージ高さの上端の数値を返します。
		 * 読み取り専用
		 */
		public function get top():Number { return -getMarginTop( stage )/*ExDocument.top*/; }
		
		
		/*======================================================================*//**
		 * StageUtil.getMarginTopメソッドの代用
		 * 
		 * <p>ステージの上マージンを取得します。</p>
		 * <p>Get the top margin of the stage.</p>
		 * 
		 * @langversion 3.0
		 * @playerversion Flash 9.0.45.0
		 * 
		 * @param stage
		 * 	<p>マージンを取得したい stage インスタンスです。</p>
		 * 	<p>The stage instance to get the margin.</p>
		 * @return
		 * 	<p>上マージンです。</p>
		 * 	<p>The top margin.</p>
		 * 
		 * @example <listing version="3.0" >
		 * </listing>
		 */
		public static function getMarginTop( stage:Stage ):Number {
			// ドキュメントクラスを取得する
			var root:Sprite = StageUtil.getDocument( stage );
			
			try {
				var rootHeight:Number = root.loaderInfo.height;
				var stgHeight:Number = stage.stageHeight;
			}
			catch ( e:Error ) {
				return -1;
			}
			
			switch ( stage.scaleMode ) {
				case StageScaleMode.NO_SCALE	: {
					switch ( stage.align ) {
						case StageAlign.TOP				:
						case StageAlign.TOP_LEFT		:
						case StageAlign.TOP_RIGHT		: { return 0; }
						case StageAlign.BOTTOM			:
						case StageAlign.BOTTOM_LEFT		:
						case StageAlign.BOTTOM_RIGHT	: { return ( stgHeight - rootHeight ); }
						default							: { return ( stgHeight - rootHeight ) / 2; }
					}
				}
				case StageScaleMode.EXACT_FIT	:
				case StageScaleMode.NO_BORDER	:
				case StageScaleMode.SHOW_ALL	: { return 0; }
			}
			
			return 0;
		}
		
		
		/**
		 * ステージ高さの中心の数値を返します。
		 * 読み取り専用
		 */
		public function get middle():Number { return centerY; }
		
		/**
		 * ステージ高さの中心の数値を返します。
		 * 読み取り専用
		 */
		public function get centerY():Number { return bottom / 2 + top / 2; }
		
		/**
		 * ステージ高さの下端の数値を返します。
		 * 読み取り専用
		 */
		public function get bottom():Number { return top + stageHeight; }
		
		/**
		 * ステージ幅の左端の数値を返します。
		 * 読み取り専用
		 */
		public function get width():Number { return stageWidth; }
		
		/**
		 * ステージ幅の左端の数値を返します。
		 * 読み取り専用
		 */
		public function get height():Number { return stageHeight; }
		
		/**
		 * ドキュメント規定幅とステージ幅の現在の倍率を返します。
		 */
		public function get scaleX():Number { return stageWidth / contentWidth; }
		
		/**
		 * ドキュメント規定高さとステージ高さの現在の倍率を返します。
		 */
		public function get scaleY():Number { return stageHeight / contentHeight; }
		
		/**
		 * LiquidLayout がイベントを受け取るか否かを取得もしくは設定します。
		 */
		public function get eventHandlerEnabled():Boolean { return _eventHandlerEnabled; }
		public function set eventHandlerEnabled( value:Boolean ):void
		{
			if ( _eventHandlerEnabled = value )
			{
				addExclusivelyEventListener( Event.RESIZE, _resizeHandler, false, 0, true );
				addExclusivelyEventListener( DocumentEvent.RESIZE_START, _resizeStartHandler, false, 0, true );
				addExclusivelyEventListener( DocumentEvent.RESIZE_PROGRESS, _resizeProgressHandler, false, 0, true );
				addExclusivelyEventListener( DocumentEvent.RESIZE_COMPLETE, _resizeCompleteHandler, false, 0, true );
			}
			else
			{
				completelyRemoveEventListener( Event.RESIZE, _resizeHandler );
				completelyRemoveEventListener( DocumentEvent.RESIZE_START, _resizeStartHandler );
				completelyRemoveEventListener( DocumentEvent.RESIZE_PROGRESS, _resizeProgressHandler );
				completelyRemoveEventListener( DocumentEvent.RESIZE_COMPLETE, _resizeCompleteHandler );
			}
		}
		private var _eventHandlerEnabled:Boolean = false;
		
		
		/**
		 * 特別プロパティのクラスを登録します。
		 * @param	propertyName	対応プロパティ名
		 * @param	specialClass	プロパティ適応クラス
		 */
		public static function registSpecialProperties( propertyName:String, specialClass:Class ):void
		{
			_specialProperties[propertyName] = specialClass;
		}
		
		/**
		 * 新しいレイアウトを作成します。
		 * @param	document		ExDocumentオブジェクト
		 *
		 */
		public function LiquidLayout( document:ExDocument, minWidth:int = 0, minHeight:int = 0, maxWidth:int = 2880, maxHeight:int = 2880 )
		{
			_document = document;
			
			_stageMinWidth  = minWidth;
			_stageMinHeight = minHeight;
			_stageMaxWidth  = maxWidth;
			_stageMaxHeight = maxHeight;
			
			_listeners = new Dictionary( true );
			
			// ドキュメントのロードが完了していない場合はロードが完了するのを待ちます。
			if ( ExDocument.loaderInfo == null )
			{
				_document.addExclusivelyEventListener( DocumentEvent.INIT, _init, false, int.MAX_VALUE, true );
			}
			else
			{
				_initialize();
			}
		}
		
		/**
		 * 初期化する
		 */
		private function _initialize():void
		{
			if ( ExDocument.loaderInfo == null )
			{
				if ( !_document.hasEventListener( DocumentEvent.INIT ) )
				{
					_document.addExclusivelyEventListener( DocumentEvent.INIT, _init, false, int.MAX_VALUE, true );
				}
				return;
			}
			
			_initialized = true;
			
			eventHandlerEnabled = true;
			
			_document.completelyRemoveEventListener( DocumentEvent.INIT, _init );
			
			stage.addEventListener( Event.RESIZE, dispatchEvent, false, 0, true );
			_document.addExclusivelyEventListener( DocumentEvent.RESIZE_START, dispatchEvent, false, 0, true );
			_document.addExclusivelyEventListener( DocumentEvent.RESIZE_PROGRESS, dispatchEvent, false, 0, true );
			_document.addExclusivelyEventListener( DocumentEvent.RESIZE_COMPLETE, dispatchEvent, false, 0, true );

		}
		
		/**
		 * ドキュメントがロード完了した際に送出されます。
		 * @param	e
		 */
		private function _init( e:DocumentEvent ):void
		{
			_document.completelyRemoveEventListener( DocumentEvent.INIT, _init );
			
			// 初期化する
			_initialize();
		}
		
		public function destroy():void
		{
			eventHandlerEnabled = false;
			_initialized = false;
			_document.completelyRemoveEventListener( DocumentEvent.INIT, _init );
			stage.removeEventListener( Event.RESIZE, dispatchEvent );
			_document.completelyRemoveEventListener( DocumentEvent.RESIZE_START, dispatchEvent );
			_document.completelyRemoveEventListener( DocumentEvent.RESIZE_PROGRESS, dispatchEvent );
			_document.completelyRemoveEventListener( DocumentEvent.RESIZE_COMPLETE, dispatchEvent );
			
			_document = null;
			
			removeAllLayouts();
			
			_listeners = null;
		}
		
		/**
		 * レイアウト制御させるオブジェクトを追加します。
		 * リサイズされた際および登録された際に変更を加えたいオブジェクトと、変更するプロパティと値をオブジェクトで設定します。
		 * プロパティオブジェクトconfigには各Layoutプロパティおよびイベントに関する記述が含まれます。
		 * 使用可能なプロパティキー一覧
		 * <ul>
		 * <li>left</li>
		 * <li>center</li>
		 * <li>right</li>
		 * <li>top</li>
		 * <li>middle</li>
		 * <li>bottom</li>
		 * <li>centerX</li>
		 * <li>centerY</li>
		 * <li>width</li>
		 * <li>height</li>
		 * <li>stageWidth</li>
		 * <li>stageHeight</li>
		 * <li>scaleX</li>
		 * <li>scaleY</li>
		 * 
		 * <li>leftPropName</li>
		 * <li>centerPropName</li>
		 * <li>rightPropName</li>
		 * <li>topPropName</li>
		 * <li>middlePropName</li>
		 * <li>bottomPropName</li>
		 * <li>centerXPropName</li>
		 * <li>centerYPropName</li>
		 * <li>widthPropName</li>
		 * <li>heightPropName</li>
		 * <li>stageWidthPropName</li>
		 * <li>stageHeightPropName</li>
		 * <li>scaleXPropName</li>
		 * <li>scaleYPropName</li>
		 * 
		 * <li>eventType</li>
		 * 
		 * <li>onResize</li>
		 * <li>onResizeParam</li>
		 * <li>onResizeStart</li>
		 * <li>onResizeStartParam</li>
		 * <li>onResizeProgress</li>
		 * <li>onResizeProgressParam</li>
		 * <li>onResizeComplete</li>
		 * <li>onResizeCompleteParam</li>
		 * </ul>
		 *
		 * デフォルトではleft,center,rightはそれぞれxプロパティ、top,middle,bottomはそれぞれyプロパティ、width,height,scaleX,scaleYはそれぞれ同名のプロパティに作用します。
		 * もし他のプロパティに作用させたい場合は、**Propertyに作用させたいプロパティ名を文字列で追加します。
		 * valueに文字列で代入するとパーセント仕様になります。width: "50" は50%の幅を意味します。
		 *
		 * eventTypeは適応するイベントのタイミングを示します。
		 * デフォルトはLayoutEventType.EVENT_ALWAYSです。
		 *
		 * on**はそれぞれのイベントのタイミングで実行させたい関数および引数を設定します。
		 * on**Paramは引数を含む配列です。
		 *
		 * @example <listing version="3.0" >
		 *
		 * </listing>
		 *
		 * @param	target
		 * @param	config
		 */
		public function addLayout( target:Object, config:Object ):void
		{
			
			var listener:LiquidLayoutListener = getLayoutListenerByTarget( target );
			if ( listener == null )
			{
				listener = new LiquidLayoutListener(  );
			}
			
			listener.eventType = ( config.eventType && _restrictedEvents[config.eventType] ) ? config.eventType : listener.eventType;
			
			if ( config.enabled === false )
			{
				listener.enabled = false;
			}
			else if ( target is DisplayObject )
			{
				if ( target.stage == null )
				{
					listener.enabled = false;
					target.addEventListener( Event.ADDED_TO_STAGE, _targetAddedToStageHandler );
				}
				else
				{
					listener.enabled = true;
					target.addEventListener( Event.REMOVED_FROM_STAGE, _targetRemovedToStageHandler );
				}
			}
			else
			{
				listener.enabled = true;
			}
			
			listener.onResize               ||= config.onResize;
			listener.onResizeParams         ||= config.onResizeParams;
			listener.onResizeStart          ||= config.onResizeStart;
			listener.onResizeStartParams    ||= config.onResizeStartParams;
			listener.onResizeProgress       ||= config.onResizeProgress;
			listener.onResizeProgressParams ||= config.onResizeProgressParams;
			listener.onResizeComplete       ||= config.onResizeComplete;
			listener.onResizeCompleteParams ||= config.onResizeCompleteParams;
			
			for ( var p:String in config )
			{
				if ( !_restrictedWords[p] )
				{
					if( _specialProperties[p] )
					{
						var propClass:Class = _specialProperties[p] as Class;
						if ( propClass == null)
						{
							continue;
						}
						try
						{
							var lp:LiquidLayoutProperty = LiquidLayoutProperty( new propClass( p, config[p], p ));
						}
						catch( e:Error )
						{
							continue;
						}
						
						listener.addProperty( lp );
					}
					else if ( _correspondingProperties[p] )
					{
						var propName:String = ( config[p + SUB_PROPERTY_NAME_SUFFIX] ) ? config[p + SUB_PROPERTY_NAME_SUFFIX] : _correspondingProperties[p];
						listener.addProperty( new LiquidLayoutProperty( p, config[p], propName ) );
					}
				}
			}
			addLayoutListener( target, listener );
		}
		
		/**
		 * LiquidLayoutListener を追加します。
		 * @param	listener
		 */
		public function addLayoutListener( target:Object, listener:LiquidLayoutListener ):void
		{
			_listeners[ target ] = listener;
			
			updateLayoutListenerAt( target, listener );
		}
		
		/**
		 * リスナー登録されている target オブジェクトから LiquidLayoutListener インスタンスを返します。
		 * @param	target
		 * @return
		 */
		public function getLayoutListenerByTarget( target:Object ):LiquidLayoutListener
		{
			return _listeners[ target ] as LiquidLayoutListener;
		}
		
		/**
		 * 全ての登録されているオブジェクトに更新を試行します。
		 * イベントタイプを指定しない場合強制的にレイアウト制御を実行します
		 * @param	eventType
		 */
		private function updateLayouts( eventType:String = null ):void
		{
			for ( var target:Object in _listeners )
			{
				updateLayoutListenerAt( target, _listeners[ target ], eventType );
			}
		}
		
		/**
		 * オブジェクトを更新します。
		 * @param	listener
		 * @param	eventType
		 * @return
		 */
		private function updateLayoutListenerAt( target:Object, listener:LiquidLayoutListener, eventType:String = null ):void
		{
			if ( !_initialized ) return;
			if ( listener == null ) return;
			
			try
			{
				listener.update( this, target, eventType );
			}
			catch ( e:Error )
			{
				errorHandler( e );
			}
			return;
		}
		
		/**
		 * 強制的に全ての登録レイアウトのレイアウト制御を実行します。
		 * ステージイベントの発行されないタイミングなどで制御を実行したい場合に使用します。
		 */
		public function updateAll():void
		{
			updateLayouts();
		}
		
		/**
		 * 強制的に指定したターゲットのレイアウト制御を実行します。
		 * ステージイベントの発行されないタイミングなどで制御を実行したい場合に使用します。
		 * @param	target	実行するターゲットもしくはそれらを含む配列
		 */
		public function updateAt( ... targets ):void
		{
			for ( var i:uint = 0; i < targets.length; i++ )
			{
				var listener:LiquidLayoutListener = getLayoutListenerByTarget( targets[i] );
				if ( listener )
				{
					updateLayoutListenerAt( targets[i], listener );
				}
			}
		}
		
		/**
		 * 無効にされているレイアウト制御を有効にします。
		 * @param	target	有効にするターゲットもしくはそれらを含む配列
		 */
		public function setEnableLayout( target:Object, enabled:Boolean ):void
		{
			var listener:LiquidLayoutListener = getLayoutListenerByTarget( target );
			if ( listener == null ) { return; }
			if ( listener.enabled = enabled )
			{
				updateLayoutListenerAt( target, listener );
			}
		}
		
		/**
		 * targetがレイアウト制御が有効かどうかを返します。
		 * @param	target	確認するターゲット
		 * @return			有効ならtrue、無効または登録がなければfalseを返します。
		 */
		public function isEnabled( target:Object ):Boolean
		{
			var d:LiquidLayoutListener = _listeners[target];
			if ( d == null ) return false;
			return d.enabled;
		}
		
		/**
		 * ターゲットを指定してレイアウト制御リストから削除します。
		 * @param	target	削除するターゲットもしくはそれらを含む配列
		 */
		public function removeLayout( ... targets ):void
		{
			for ( var i:uint = 0; i < targets.length; i++ )
			{
				removeLayoutAt( targets[i] );
			}
		}
		
		/**
		 * レイアウト制御リストから削除します。
		 * @param	target	削除するターゲットもしくはそれらを含む配列
		 */
		public function removeLayoutAt( target:Object ):void
		{
			var listener:LiquidLayoutListener = getLayoutListenerByTarget( target );
			if ( listener )
			{
				listener.destroy();
			}
			delete _listeners[ target ];
			listener = null;
		}
		
		/**
		 * すべてのレイアウト制御を削除します。
		 */
		public function removeAllLayouts():void
		{
			for( var target:Object in _listeners )
			{
				removeLayoutAt( target );
			}
		}
		
		/**
		 * リサイズイベント
		 */
		private function _resizeHandler( e:Event ):void
		{
			updateLayouts( LiquidLayoutEvent.RESIZE_ALWAYS );
			dispatchEvent( new LiquidLayoutEvent( LiquidLayoutEvent.RESIZE_ALWAYS, false, false, this ));
		}
		
		/**
		 * ステージのリサイズを開始した場合に送出されます。
		 */
		private function _resizeStartHandler( e:Event ):void
		{
			updateLayouts( LiquidLayoutEvent.RESIZE_START );
		}
		
		/**
		 * ステージがリサイズ実行中に送出されます。
		 */
		private function _resizeProgressHandler( e:Event ):void
		{
			updateLayouts( LiquidLayoutEvent.RESIZE_PROGRESS );
		}
		
		/**
		 * ステージのリサイズが終了した場合に送出されます。
		 */
		private function _resizeCompleteHandler( e:Event ):void
		{
			updateLayouts( LiquidLayoutEvent.RESIZE_COMPLETE );
			
		}
		
		/**
		 * 追加されたときにオブジェクトがステージ上に登録されてなかったときに送出。
		 */
		private function _targetAddedToStageHandler( e:Event ):void
		{
			e.target.removeEventListener( Event.ADDED_TO_STAGE, _targetAddedToStageHandler );
			setEnableLayout( e.target, true );
			e.target.addEventListener( Event.REMOVED_FROM_STAGE, _targetRemovedToStageHandler );
		}
		
		/**
		 * オブジェクトがステージ上に登録しなおされた場合に送出
		 */
		private function _targetRemovedToStageHandler( e:Event ):void
		{
			setEnableLayout( e.target, false );
		}
		
		/**
		 * エラーハンドラ
		 * 適当スグル・・・
		 */
		private function errorHandler( err:Error ):void
		{
			trace( err.message );
		}
	}
	
}